#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <cstdlib>
#include <iostream>
#include <time.h>
#define BUFSIZE 40
#define TIMEOUT_S 1
#define TIMEOUT_U 0
#define DEBUG 1

enum TTY_ERROR
{
    TTY_OK = 0,
    TTY_READ_ERROR = -1,
    TTY_WRITE_ERROR = -2,
    TTY_SELECT_ERROR = -3,
    TTY_TIME_OUT = -4,
    TTY_PORT_FAILURE = -5,
    TTY_PARAM_ERROR = -6,
    TTY_ERRNO = -7,
    TTY_OVERFLOW = -8,
    TTY_PORT_BUSY = -9,
};


using namespace std;

char ser_i_buf[BUFSIZE];
char ser_o_buf[BUFSIZE];

int
set_interface_attribs(int fd, int speed, int parity)
{
    struct termios tty;
    if (tcgetattr(fd, &tty) != 0)
    {
        return -1;
    }

    cfsetospeed(&tty, speed);
    cfsetispeed(&tty, speed);

    tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS8;     // 8-bit chars
    // disable IGNBRK for mismatched speed tests; otherwise receive break
    // as \000 chars
    tty.c_iflag &= ~IGNBRK;         // disable break processing
    tty.c_lflag = 0;                // no signaling chars, no echo,
    // no canonical processing
    tty.c_oflag = 0;                // no remapping, no delays
    tty.c_cc[VMIN] = 0;            // read doesn't block
    tty.c_cc[VTIME] = 30;            // 0.5 seconds read timeout

    tty.c_iflag &= ~(IXON | IXOFF | IXANY); // shut off xon/xoff ctrl

    tty.c_cflag |= (CLOCAL | CREAD);// ignore modem controls,
    // enable reading
    tty.c_cflag &= ~(PARENB | PARODD);      // shut off parity
    tty.c_cflag |= parity;
    tty.c_cflag &= ~CSTOPB;
    tty.c_cflag &= ~CRTSCTS;

    if (tcsetattr(fd, TCSANOW, &tty) != 0)
    {
        return -1;
    }
    return 0;
}

int tty_timeout(int fd, int timeout_s, int timeout_u)
{
    if (fd == -1)
        return TTY_ERRNO;

    struct timeval tv;
    fd_set readout;
    int retval;

    FD_ZERO(&readout);
    FD_SET(fd, &readout);

    /* wait for 'timeout' seconds */
    tv.tv_sec = timeout_s;
    tv.tv_usec = timeout_u;

    /* Wait till we have a change in the fd status */
    retval = select(fd + 1, &readout, nullptr, nullptr, &tv);

    if (retval > 0)
        return TTY_OK;
    else if (retval == -1)
        return TTY_SELECT_ERROR;
    else
        return TTY_TIME_OUT;

}

void
set_blocking(int fd, int should_block)
{
    struct termios tty;
    memset(&tty, 0, sizeof tty);
    if (tcgetattr(fd, &tty) != 0)
    {
        return;
    }

    tty.c_cc[VMIN] = should_block ? 1 : 0;
    tty.c_cc[VTIME] = 1;            // 0.5 seconds read timeout

}

int ttysend(int fd, char *str, int *nbytes_written)
{
    if (fd == -1)
        return TTY_ERRNO;

    int nbytes = strlen(str);
    int bytes_w = 0;
    *nbytes_written = 0;

    while (nbytes > 0)
    {
        bytes_w = write(fd, str + (*nbytes_written), nbytes);

        if (bytes_w < 0)
            return TTY_WRITE_ERROR;

        *nbytes_written += bytes_w;
        nbytes -= bytes_w;
    }

    return TTY_OK;
}

int ttyrecv(int fd, char *buf, int nbytes, int *nbytes_read)
{

    if (fd == -1)
        return TTY_ERRNO;

    int numBytesToRead = nbytes;
    int bytesRead = 0;
    int err = 0;
    *nbytes_read = 0;
    char *buffer = buf;

    if (nbytes <= 0)
        return TTY_PARAM_ERROR;

    while (numBytesToRead > 0)
    {

        if ((err = tty_timeout(fd, TIMEOUT_S, TIMEOUT_U)))
        {
            return err;
        }

        bytesRead = read(fd, buffer + (*nbytes_read), ((uint32_t) numBytesToRead));

        if (bytesRead < 0)
            return TTY_READ_ERROR;

        *nbytes_read += bytesRead;
        numBytesToRead -= bytesRead;

        // Look for the Meade termination character.
        if (buffer[(*nbytes_read) - 1] == '#')
        {
            break;
        }
    }
    buf[*nbytes_read] = '\0';
    return TTY_OK;

}

int open_port(char *port_name)
{

    int fd = open(port_name, O_RDWR | O_NOCTTY | O_NONBLOCK);
    if (fd < 0)
    {
        return fd;
    }

    set_interface_attribs(fd, B9600, 0);  // set speed to 9600 bps, 8n1 (no parity)
    set_blocking(fd, 0);                // set no blocking
    return fd;
}

int send_set_time(int fd) {

    char i_buf[BUFSIZE];
    char o_buf[BUFSIZE];

    int wr_len;
    int rd_len;
    time_t rawtime;
    struct tm * timeinfo;
    
    time (&rawtime);
    timeinfo = localtime (&rawtime);

    strftime (o_buf,BUFSIZE,":hI%y%m%d%H%M%S#",timeinfo);
    memset(i_buf, 0, sizeof(i_buf));
    if (DEBUG) sleep(3);
    wr_len = 0;
    rd_len = 0;

    ttysend(fd, o_buf, &wr_len);
    cout << " > [" << o_buf << "]" << endl;
    ttyrecv(fd, i_buf, BUFSIZE, &rd_len);
    cout << " < [" << i_buf << "]" << endl;
    if (i_buf[0] == '1') {
        return (0);
    } else {
        return (1);
    }
}


int main(int argc, char **argv) {
    char csv_name[256];
    char port_name[64];

    if (argc == 3) {
        if (!strcmp(argv[1], "time")) {
            strncpy(port_name, argv[2], sizeof(port_name) - 1);
            cout << port_name << endl;
            return send_set_time(open_port(port_name));
        }
    } else {
        cout << "lx_init time <port>" << endl;
    }
    return 0;
}

